//
//  FontRepresentable+Extension.swift
//
//
//  Created by Divyesh Vekariya on 26/12/23.
//

import SwiftUI

extension FontRepresentable {
    /// Check if font is bold
    var isBold: Bool {
        return fontDescriptor.symbolicTraits.contains(.bold)
    }
    
    /// Check if font is italic
    var isItalic: Bool {
        return fontDescriptor.symbolicTraits.contains(.italic)
    }
    
    /// Make font **Bold**
    func makeBold() -> FontRepresentable {
        if isBold {
            return self
        } else {
            let fontDesc = fontDescriptor.withSymbolicTraits(.bold)
            return FontRepresentable(descriptor: fontDesc, size: pointSize)!
        }
    }
    
    /// Make font **Italic**
    func makeItalic() -> FontRepresentable {
        if isItalic {
            return self
        } else {
            let fontDesc = fontDescriptor.withSymbolicTraits(.italic)
            return FontRepresentable(descriptor: fontDesc, size: pointSize)!
        }
    }
    
    /// Make font **Bold** and **Italic**
    func setBoldItalicStyles() -> FontRepresentable {
        return makeBold().makeItalic()
    }
    
    /// Remove **Bold** style from font
    func removeBoldStyle() -> FontRepresentable {
        if !isBold {
            return self
        } else {
            let fontDesc = fontDescriptor.withSymbolicTraits(fontDescriptor.symbolicTraits.subtracting(.bold))
            return FontRepresentable(descriptor: fontDesc, size: pointSize)!
        }
    }
    
    /// Remove **Italic** style from font
    func removeItalicStyle() -> FontRepresentable {
        if !isItalic {
            return self
        } else {
            let fontDesc = fontDescriptor.withSymbolicTraits(fontDescriptor.symbolicTraits.subtracting(.italic))
            return FontRepresentable(descriptor: fontDesc, size: pointSize)!
        }
    }
    
    /// Remove **Bold** and **Italic** style from font
    func makeNormal() -> FontRepresentable {
        return removeBoldStyle().removeItalicStyle()
    }
    
    /// Toggle **Bold** style of font
    func toggleBoldTrait() -> FontRepresentable {
        if isBold {
            return removeBoldStyle()
        } else {
            return makeBold()
        }
    }
    
    /// Toggle **Italic** style of font
    func toggleItalicStyle() -> FontRepresentable {
        if isItalic {
            return removeItalicStyle()
        } else {
            return makeItalic()
        }
    }
    
    /// Get a new font with updated font size by **size**
    func updateFontSize(size: CGFloat) -> FontRepresentable {
        if pointSize != size {
            return FontRepresentable(descriptor: fontDescriptor, size: size)!
        } else {
            return self
        }
    }
    
    func updateFontSize(multiple: CGFloat) -> FontRepresentable {
        if pointSize != multiple * pointSize {
            let size = multiple * pointSize
            return FontRepresentable(descriptor: fontDescriptor, size: size)!
        } else {
            return self
        }
    }
}

public extension FontRepresentable {
    /// Get a new font by adding a text style.
    func addFontStyle(_ style: TextSpanStyle) -> FontRepresentable {
        guard let trait = style.symbolicTraits, !fontDescriptor.symbolicTraits.contains(trait) else { return self }
        let fontDesc = fontDescriptor.withSymbolicTraits(fontDescriptor.symbolicTraits.union(trait))
        return FontRepresentable(descriptor: fontDesc, size: pointSize)!
    }
    
    ///Get a new font by removing a text style.
    func removeFontStyle(_ style: TextSpanStyle) -> FontRepresentable {
        guard let trait = style.symbolicTraits, fontDescriptor.symbolicTraits.contains(trait) else { return self }
        let fontDesc = fontDescriptor.withSymbolicTraits(fontDescriptor.symbolicTraits.subtracting(trait))
        return FontRepresentable(descriptor: fontDesc, size: pointSize)!
    }
    
    /// Get a new font by toggling a text style.
    func byTogglingFontStyle(_ style: TextSpanStyle) -> FontRepresentable {
        guard let trait = style.symbolicTraits else { return self }
        let newTraits = fontDescriptor.symbolicTraits.symmetricDifference(trait)
        let fontDesc = fontDescriptor.withSymbolicTraits(newTraits)
        return FontRepresentable(descriptor: fontDesc, size: pointSize)!
    }
}
