//
//  TextViewWrapper.swift
//
//
//  Created by Divyesh Vekariya on 12/10/23.
//

import SwiftUI
import AppKit

struct TextViewWrapper: NSViewRepresentable {
    @ObservedObject var state: RichEditorState
    
    @Binding private var typingAttributes: [NSAttributedString.Key: Any]?
    @Binding private var attributesToApply: ((spans: [(span: RichTextSpanInternal, shouldApply: Bool)], onCompletion: () -> Void))?
    
    private let isEditable: Bool
    private let isUserInteractionEnabled: Bool
    private let isScrollEnabled: Bool
    private let linelimit: Int?
    private let fontStyle: NSFont?
    private let fontColor: NSColor
    private let backGroundColor: NSColor
    private let tag: Int?
    private let onTextViewEvent: ((TextViewEvents) -> Void)?
    
    init(state: ObservedObject<RichEditorState>,
         typingAttributes: Binding<[NSAttributedString.Key: Any]?>? = nil,
         attributesToApply: Binding<(spans: [(span: RichTextSpanInternal, shouldApply: Bool)], onCompletion: () -> Void)?>? = nil,
         isEditable: Bool = true,
         isUserInteractionEnabled: Bool = true,
         isScrollEnabled: Bool = false,
         linelimit: Int? = nil,
         fontStyle: NSFont? = nil,
         fontColor: NSColor = .black,
         backGroundColor: NSColor = .clear,
         tag: Int? = nil,
         onTextViewEvent: ((TextViewEvents) -> Void)?) {
        self._state = state
        self._typingAttributes = typingAttributes != nil ? typingAttributes! : .constant(nil)
        self._attributesToApply = attributesToApply != nil ? attributesToApply! : .constant(nil)
        
        self.isEditable = isEditable
        self.isUserInteractionEnabled = isUserInteractionEnabled
        self.isScrollEnabled = isScrollEnabled
        self.linelimit = linelimit
        self.fontStyle = fontStyle
        self.fontColor = fontColor
        self.backGroundColor = backGroundColor
        self.tag = tag
        self.onTextViewEvent = onTextViewEvent
    }
    
    func makeNSView(context: Context) -> NSTextView {
        let textView = NSTextView()
        textView.delegate = context.coordinator
        textView.isEditable = isEditable
        textView.isSelectable = isUserInteractionEnabled
        textView.drawsBackground = false
        textView.backgroundColor = .clear
        
        if let fontStyle {
            textView.typingAttributes = [.font: fontStyle]
        }
        
        textView.textStorage?.setAttributedString(state.editableText)
        
        textView.textColor = fontColor
        textView.textContainer?.lineFragmentPadding = 0
        
        if let attributes = typingAttributes {
            textView.typingAttributes = attributes
        }
        
        // Removed the tag-setting code

//        
//        if let tag = tag {
//            textView.tag = tag
//        }
        
        if let linelimit = linelimit {
            textView.textContainer?.maximumNumberOfLines = linelimit
        }
        
        return textView
    }
    
    func updateNSView(_ textView: NSTextView, context: Context) {
        textView.textColor = fontColor
        
        var attributes = getTypingAttributesForStyles(state.activeStyles)
        if !attributes.contains(where: { $0.key == .font }) {
            attributes[.font] = fontStyle
        }
        textView.typingAttributes = attributes
        
        if let data = attributesToApply {
            applyAttributesToSelectedRange(textView, spans: data.spans, onCompletion: data.onCompletion)
        }
    }
    
    func makeCoordinator() -> Coordinator {
        Coordinator(self)
    }
    
    class Coordinator: NSObject, NSTextViewDelegate {
        var parent: TextViewWrapper
        
        init(_ textView: TextViewWrapper) {
            self.parent = textView
        }
        
        func textView(_ textView: NSTextView, shouldChangeTextIn affectedCharRange: NSRange, replacementString: String?) -> Bool {
            return true
        }
        
        func textViewDidChangeSelection(_ notification: Notification) {
            guard let textView = notification.object as? NSTextView else { return }
            parent.onTextViewEvent?(.didChangeSelection(textView))
        }
        
        func textDidChange(_ notification: Notification) {
                guard let textView = notification.object as? NSTextView else { return }
                
                // Check if there's an ongoing input method editing
                if textView.hasMarkedText() == false {
                    parent.state.editableText = NSMutableAttributedString(attributedString: textView.attributedString())
                }
                parent.onTextViewEvent?(.didChange(textView))
            }
        
        func textDidBeginEditing(_ notification: Notification) {
            guard let textView = notification.object as? NSTextView else { return }
            parent.onTextViewEvent?(.didBeginEditing(textView))
        }
        
        func textDidEndEditing(_ notification: Notification) {
            guard let textView = notification.object as? NSTextView else { return }
            parent.onTextViewEvent?(.didEndEditing(textView))
        }
    }
    
    func getTypingAttributesForStyles(_ styles: Set<RichTextStyle>) -> [NSAttributedString.Key: Any] {
        var font = fontStyle
        var attributes: [NSAttributedString.Key: Any] = [:]
        
        Set(styles).forEach {
            if $0.attributedStringKey == .font {
                font = $0.getFontWithUpdating(font: font ?? .systemFont(ofSize: NSFont.systemFontSize))
                attributes[$0.attributedStringKey] = font
            } else {
                attributes[$0.attributedStringKey] = $0.defaultAttributeValue(font: fontStyle)
            }
        }
        return attributes
    }
    
    func applyAttributesToSelectedRange(_ textView: NSTextView, spans: [(span: RichTextSpanInternal, shouldApply: Bool)], onCompletion: (() -> Void)? = nil) {
        var spansToUpdate = spans.filter { $0.span.attributes?.header != nil }
        spansToUpdate.append(contentsOf: spans.filter { $0.span.attributes?.header == nil })
        spansToUpdate.forEach { span in
            span.span.attributes?.styles().forEach { style in
                textView.textStorage?.setRichTextStyle(style, to: span.shouldApply, at: span.span.spanRange)
            }
        }
        onCompletion?()
    }
}

enum TextViewEvents {
    case didChangeSelection(_ textView: NSTextView)
    case didBeginEditing(_ textView: NSTextView)
    case didChange(_ textView: NSTextView)
    case didEndEditing(_ textView: NSTextView)
}
